/*
 * Copyright 2021 - All rights reserved.
 * Rodrigo Tapia-McClung
 *
 * February-June 2021
 */

/* global mapboxgl, turf */

const windowUrl = new URL(window.location.href);
const baseUrl = windowUrl.hostname == "localhost" ?
    `${windowUrl.origin}` : // returns "http://localhost:8092"
    `${windowUrl.origin}/aire-cdmx`; // returns "http://servername/aire-cdmx"
    // use `${baseUrl}/something`

let cdmx, estaciones, ozono, fechas,
    slider = document.getElementById("slider"),
    datetimeLabel = document.getElementById("datetime"),
    playBtn = document.getElementById("play-button"),
    playing = false,
    animate;

fetch(`${baseUrl}/data/cdmx.geojson`)
    .then(response => response.json())
    .then(d => cdmx = d);

fetch(`${baseUrl}/data/estaciones.geojson`)
    .then(response => response.json())
    .then(d => estaciones = d);


const papaPromise = url => {
    return new Promise(function (resolve, reject) {
        Papa.parse(url, {
            download: true,
            header: true,
            skipEmptyLines: true,
            complete: resolve,
        });
    });
}

let ozonoPromise = papaPromise(`${baseUrl}/data/ozono.csv`); // ozono es en ppb: 10^12
// FIXME: check and fix date order in data
/*const getOzono = fetch(`${baseUrl}/ozono.csv`)
    .then(response => response.text())
    .then(d => Papa.parse(d))
    .catch(err => console.log(err))

getOzono.then(d => ozono = d.data);
*/

let map = new mapboxgl.Map({
    "container": "mexmap",
    "accessToken": "pk.eyJ1IjoiZGV2ZWxvcGdlbyIsImEiOiJja2dwcXFic20wYnJnMzBrbG11d3dwYTkyIn0.4WwFOH6C7hDQXV9obU6mAw",
    "style": "mapbox://styles/mapbox/dark-v10",
    "center": [-99.17, 19.36],
    "zoom": 10,
    "minZoom": 9,
    "maxZoom": 14
});

map.addControl(new mapboxgl.NavigationControl());

// map language - doesn't quite work
/*map.addControl(new MapboxLanguage({
    defaultLanguage: 'es'
}));*/

// TODO: display more friendly dates
const dateTimeOptions = {
    weekday: "short",
    day: "numeric",
    month:"short",
    year: "numeric",
    hour: 'numeric',
    minute: 'numeric',
    //hour: '2-digit',
    //minute: '2-digit',
    hour12: false,
    timeZone: 'America/Mexico_City'
};

//let myDate = "1577919600000";
//let myDate = "1605218400000";

const range = (start, stop, step = 1) =>
    Array(Math.ceil((stop - start) / step)).fill(start).map((x, y) => x + y * step);

const intersect = (fc1, fc2) => {
    let fc = [];
    fc1.features.forEach( f1 => {
        fc2.features.forEach( f2 => {
            let intersection = turf.intersect(f1, f2, {properties: {value: f1.properties.value}});
            if (intersection != undefined) {
                fc.push(intersection);
            }
        });
    });
    return turf.featureCollection(fc);
}

const makeSurface = date => {
    //map.getSource("estaciones").serialize().data
    // filter data for given date value
    let subdata = ozono.map(d => { return {time: d["time"], [date]: d[date]} });
    let maxValue = Math.max.apply(Math, subdata.map( d=> d[date]));
    // assign corresponding value to each station
    turf.featureEach(estaciones, point => {
        let stationData = subdata.filter( i => i.time == point.properties.cve_estac);
        point.properties.calidad = stationData[0] && parseFloat(stationData[0][date]) != -99 ? parseFloat(stationData[0][date]) : 0;
        // use value to style station dot
        map.setFeatureState(
            {
                // source tileset and source layer
                source: "estaciones",
                sourceLayer: "estaciones",
                // unique ID row name
                id: point.properties.cve_estac
            },
            // Add rows you want to style/interact with
            {
                calidad: point.properties.calidad
                //candidate: row.candidate,
            }
        );
    });

    let options = {gridType: "point", property: "calidad", units: "kilometers", weight: 2};
    let grid = turf.interpolate(estaciones, 0.75, options);

    let breaks = range(0, maxValue != -99 ? maxValue : 0 + 1, 5);
    //let lines = turf.isolines(grid, breaks, {zProperty: "calidad"});
    let bValues = breaks.map( b => { return {value: b} });
    let bands = turf.isobands(grid, breaks, {zProperty: "calidad", breaksProperties: bValues});

    //map.getSource("interpolation").setData(lines);
    //map.getSource("interpolation").setData(bands);

    let intersection = intersect(bands, cdmx);
    map.getSource("interpolation").setData(intersection);
}

let dateidx = fechas ? fechas.indexOf(parseInt(slider.value)) : 0;
const run = () => {
    if (dateidx == fechas.length) {
        clearInterval(animate);
    } else {
        dateidx++;
        makeSurface(fechas[dateidx]);
        slider.value = fechas[dateidx];
        let date = new Date(fechas[dateidx]);
        //datetimeLabel.textContent = new Intl.DateTimeFormat('es-MX', dateTimeOptions).format(date);
        datetimeLabel.textContent = date;
    }
}

map.on("style.load", async () => {

    map.addSource("estaciones", {
        "type": "vector",
        "tiles": [`${baseUrl}/estaciones/mbtiles/{z}/{x}/{y}.pbf`],
        promoteId: "cve_estac"
    });

    map.addSource("interpolation", {
        type: "geojson",
        data: {
            type: "FeatureCollection",
            features: []
        }
    });

    slider.addEventListener("input", e => {
        var date = parseInt(e.target.value, 10);
        //datetimeLabel.textContent = new Intl.DateTimeFormat('es-MX', dateTimeOptions).format(date);
        datetimeLabel.textContent = new Date(parseInt(e.target.value));
        dateidx = fechas.indexOf(date);
        makeSurface(date);
    });

    playBtn.addEventListener("click", () => {
        if (!playing) {
            playBtn.innerHTML = "<i class='material-icons'>pause</i>";
            playing =  true;
            animate = setInterval( run, 500);
        } else {
            playBtn.innerHTML = "<i class='material-icons'>play_arrow</i>";
            playing = false;
            clearInterval(animate);
        }
    });
    ozonoPromise.then( results => {
        ozono = results.data;
        fechas = Object.keys(ozono[0]).slice(1,-1).map(d => parseInt(d));
        slider.min = fechas[0];
        slider.max = fechas[fechas.length-1];
        slider.step = 3600000;
        slider.value = slider.min;
        datetimeLabel.textContent = new Date(fechas[0]);
        //datetimeLabel.textContent = new Intl.DateTimeFormat('es-MX', dateTimeOptions).format(new Date(fechas[0]));
        // calculate max value from data
        let maxValue = Math.max(...ozono.map( d => Object.values(d)).flat().filter(val => !isNaN(val)));
        /*let expression = ["match", ["get", "cve_estac"]];
        results.data.forEach( row => {
            map.setFeatureState(
                {
                    // source tileset and source layer
                    source: "estaciones",
                    sourceLayer: "estaciones",
                    // unique ID row name
                    id: row.time
                },
                // Add rows you want to style/interact with
                {
                    calidad: parseFloat(row[fechas[0]]) != -99 ? parseFloat(row[fechas[0]]) : 0
                    //candidate: row.candidate,
                }
            );
            var green = (parseFloat(row[myDate]) / maxValue) * 255;
            var color = "rgba(" + 0 + ", " + green + ", " + 0 + ", 1)";
            expression.push(row["time"], color);
        });
        expression.push("rgba(0,0,0,0)");*/

        map.addSource("cdmx", {
            "type": "geojson",
            //"data": `${baseUrl}/estaciones.geojson`
            "data": cdmx,
            //promoteId: "cve_estac"
        });
        map.addLayer({
            id: "cdmx",
            type: "line",
            source: "cdmx",
            paint: {
                "line-color":"#088"
            }
        });

        map.addLayer({
            "id": "estaciones-circle",
            "type": "circle",
            "source": "estaciones",
            "source-layer": "estaciones",
            "layout": {},
            "paint": {
                "circle-radius": 4,
                //"circle-color": "#088",
                //"circle-color": expression,
                "circle-color": [
                    "interpolate",
                    ["linear"],
                    //["get", "value"],
                    ["feature-state", "calidad"],
                    0, "rgb(0,0,0)",
                    1, "rgb(0, 228, 0)",
                    51, "rgb(255,255,0)",
                    95, "rgb(255, 126, 0)",
                    135, "rgb(255, 0, 0)",
                    175, "rgb(143, 63, 151)"
                ],
                "circle-opacity": 0.5
            }
        });

        map.addLayer({
            id: "interpolation",
            type: "fill",
            source: "interpolation",
            paint: {
                "fill-color": [
                    "interpolate",
                    ["linear"],
                    ["get", "value"],
                    0, "rgb(0, 228, 0)",
                    51, "rgb(255,255,0)",
                    95, "rgb(255, 126, 0)",
                    135, "rgb(255, 0, 0)",
                    175, "rgb(143, 63, 151)"
                ],
                "fill-opacity": 0,
                //"fill-antialias": false // use contour lines or not
            }
        });

        map.addLayer({
            "id": "extrusion",
            "type": "fill-extrusion",
            "source": "interpolation",
            "paint": {
                "fill-extrusion-color": [
                    "interpolate",
                    ["linear"],
                    ["get", "value"],
                    0, "rgb(0, 228, 0)",
                    51, "rgb(255,255,0)",
                    95, "rgb(255, 126, 0)",
                    135, "rgb(255, 0, 0)",
                    175, "rgb(143, 63, 151)"
                ],
                //"fill-extrusion-height": ['*', ["get", "value"], 5],
                "fill-extrusion-height": {
                    property: "value",
                    stops: [
                        [{ zoom: 10, value: 0 }, 1],
                        [{ zoom: 10, value: maxValue }, 1],
                        [{ zoom: 14, value: 0 }, 1],
                        [{ zoom: 14, value: maxValue }, 800]
                    ]
                },
                "fill-extrusion-base": 0,
                "fill-extrusion-height-transition": {
                    duration: 500,
                    delay: 0
                },
                /*"fill-extrusion-height": [
                    "interpolate",
                    ["linear"],
                    ["zoom"],
                    10,
                    0,
                    14,
                    ["get", "value"]
                ],
                "fill-extrusion-base": [
                    "interpolate",
                    ["linear"],
                    ["zoom"],
                    10,
                    0,
                    14,
                    ["get", "value"]
                ],*/
                "fill-extrusion-opacity": 0.5
            }
        });

        makeSurface(fechas[0]);

        // Create popup
        let popup = new mapboxgl.Popup({
            closeButton: false,
            closeOnClick: false
        });

        // What happens on hover on layer
        map.on("mouseenter", "estaciones-circle", e => {
            map.getCanvas().style.cursor = "pointer";

            let coordinates = e.features[0].geometry.coordinates.slice();
            let cve_estac = e.features[0].properties.cve_estac;
            let estacion = e.features[0].properties.nom_estac;
            let calidad = e.features[0].state.calidad != undefined ? e.features[0].state.calidad : '--';

            // Set values and display popup
            popup
                .setLngLat(coordinates)
                .setHTML("Estación: " + estacion + "<br/>Clave: " + cve_estac + "<br/>Valor: " + calidad)
                .addTo(map);
        });

        // Remove popup on mouse out
        map.on("mouseleave", "estaciones-circle", () => {
            map.getCanvas().style.cursor = "";
            popup.remove();
        });

        // bring circle layer to top
        map.moveLayer("estaciones-circle");
    });
});